/// <summary>
/// Project: WinUsb_cs
/// 
/// ***********************************************************************
/// Software License Agreement
///
/// Licensor grants any person obtaining a copy of this software ("You") 
/// a worldwide, royalty-free, non-exclusive license, for the duration of 
/// the copyright, free of charge, to store and execute the Software in a 
/// computer system and to incorporate the Software or any portion of it 
/// in computer programs You write.   
/// 
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
/// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
/// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
/// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
/// THE SOFTWARE.
/// ***********************************************************************
/// 
/// Author               Date        Version
/// Jan Axelson          6/3/08     1.3
///                      8/8/08     1.3.1 Because WinUsb_QueryDeviceInformation appears
///                                       unreliable for detecting device speed, I
///                                       commented out the call to obtain device speed. 
///                      10/1/08    1.4   Minor edits     
///                      10/29/08   1.5   Minor edits mainly for 64-bit compatibility
///                      11/1/08    1.6   Minor edits  
///                      11/9/08    1.7   Minor edits    
///                      2/10/09    1.8   Changes to WinUsb_ReadPipe parameters.
///                      2/11/09    1.81  Moved Free_ and similar to Finally blocks
///                                       to ensure they execute.
/// 
/// 
/// This software was created using Visual Studio 2008 Standard Edition with .NET Framework 2.0.
/// 
/// Purpose: 
/// Demonstrates USB communications using the Microsoft WinUSB driver.
/// 
/// Requirements:
/// Windows XP or later and an attached USB device that uses the WinUSB driver.
/// 
/// Description:
/// Finds an attached device whose INF file contains a specific device interface GUID.
/// Enables sending and receiving data via bulk, interrupt, and control transfers.
/// 
/// Uses RegisterDeviceNotification() and WM_DEVICE_CHANGE messages
/// to detect when a device is attached or removed.
/// 
/// For bulk and interrupt transfers, the application uses a Delegate and the BeginInvoke 
/// and EndInvoke methods to read data asynchronously, so the application's main thread 
/// doesn't have to wait for the device to return data. A callback routine uses 
/// marshaling to send data to the form, whose code runs in a different thread. 
///  
/// This software, an example INF file, and companion device firmware are available from 
/// www.Lvr.com
/// 
/// Send comments, bug reports, etc. to jan@Lvr.com 
/// This application has been tested under Windows XP and Windows Vista.
/// </summary>

using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace WinUsbDemo
{
	/// <summary>
	/// The application's form. Buttons enable detecting a device
	/// and initiating transfers.
	/// </summary>
	/// 
	internal class frmMain : System.Windows.Forms.Form
	{
		#region '"Windows Form Designer generated code "'
		public frmMain()
			: base()
		{


			// This call is required by the Windows Form Designer.
			InitializeComponent();
		}
		// Form overrides dispose to clean up the component list.
		protected override void Dispose(Boolean Disposing)
		{
			if (Disposing)
			{
				if (!(components == null))
				{
					components.Dispose();
				}
			}
			base.Dispose(Disposing);
		}

		// Required by the Windows Form Designer
		private System.ComponentModel.IContainer components;
		public System.Windows.Forms.ToolTip ToolTip1;
		public System.Windows.Forms.Button cmdSendandReceiveViaBulkTransfers;
		public System.Windows.Forms.GroupBox fraControlReadTransfer;
		public System.Windows.Forms.TextBox txtBulkDataToSend;
		public System.Windows.Forms.GroupBox fraBulkTransfer;
		public System.Windows.Forms.ComboBox cboInterruptOutByte1;
		public System.Windows.Forms.ComboBox cboInterruptOutByte0;
		public System.Windows.Forms.GroupBox fraInterruptTransfer;
		public System.Windows.Forms.ListBox lstResults;
		// NOTE: The following procedure is required by the Windows Form Designer
		// It can be modified using the Windows Form Designer.
		// Do not modify it using the code editor.
		internal System.Windows.Forms.Button cmdFindDevice;
		internal System.Windows.Forms.Button cmdSendAndReceiveViaInterruptTransfers;
		public System.Windows.Forms.GroupBox fraControWriteTransfer;
		internal System.Windows.Forms.Button cmdControlWriteTransfer;
		public System.Windows.Forms.ComboBox cboControlWriteByte1;
		public System.Windows.Forms.ComboBox cboControlWriteByte0;
		internal System.Windows.Forms.Button cmdControlReadTransfer;

		[System.Diagnostics.DebuggerStepThrough()]
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.ToolTip1 = new System.Windows.Forms.ToolTip(this.components);
			this.fraControlReadTransfer = new System.Windows.Forms.GroupBox();
			this.cmdControlReadTransfer = new System.Windows.Forms.Button();
			this.cmdSendandReceiveViaBulkTransfers = new System.Windows.Forms.Button();
			this.fraBulkTransfer = new System.Windows.Forms.GroupBox();
			this.txtBulkDataToSend = new System.Windows.Forms.TextBox();
			this.fraInterruptTransfer = new System.Windows.Forms.GroupBox();
			this.cmdSendAndReceiveViaInterruptTransfers = new System.Windows.Forms.Button();
			this.cboInterruptOutByte1 = new System.Windows.Forms.ComboBox();
			this.cboInterruptOutByte0 = new System.Windows.Forms.ComboBox();
			this.lstResults = new System.Windows.Forms.ListBox();
			this.cmdFindDevice = new System.Windows.Forms.Button();
			this.fraControWriteTransfer = new System.Windows.Forms.GroupBox();
			this.cboControlWriteByte1 = new System.Windows.Forms.ComboBox();
			this.cboControlWriteByte0 = new System.Windows.Forms.ComboBox();
			this.cmdControlWriteTransfer = new System.Windows.Forms.Button();
			this.fraControlReadTransfer.SuspendLayout();
			this.fraBulkTransfer.SuspendLayout();
			this.fraInterruptTransfer.SuspendLayout();
			this.fraControWriteTransfer.SuspendLayout();
			this.SuspendLayout();
			// 
			// fraControlReadTransfer
			// 
			this.fraControlReadTransfer.BackColor = System.Drawing.SystemColors.Control;
			this.fraControlReadTransfer.Controls.Add(this.cmdControlReadTransfer);
			this.fraControlReadTransfer.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.fraControlReadTransfer.ForeColor = System.Drawing.SystemColors.ControlText;
			this.fraControlReadTransfer.Location = new System.Drawing.Point(237, 27);
			this.fraControlReadTransfer.Name = "fraControlReadTransfer";
			this.fraControlReadTransfer.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.fraControlReadTransfer.Size = new System.Drawing.Size(175, 94);
			this.fraControlReadTransfer.TabIndex = 7;
			this.fraControlReadTransfer.TabStop = false;
			this.fraControlReadTransfer.Text = "Control Read Transfer";
			// 
			// cmdControlReadTransfer
			// 
			this.cmdControlReadTransfer.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cmdControlReadTransfer.Location = new System.Drawing.Point(49, 28);
			this.cmdControlReadTransfer.Name = "cmdControlReadTransfer";
			this.cmdControlReadTransfer.Size = new System.Drawing.Size(71, 36);
			this.cmdControlReadTransfer.TabIndex = 12;
			this.cmdControlReadTransfer.Text = "Send";
			this.cmdControlReadTransfer.UseVisualStyleBackColor = true;
			this.cmdControlReadTransfer.Click += new System.EventHandler(this.cmdControlReadTransfer_Click);
			// 
			// cmdSendandReceiveViaBulkTransfers
			// 
			this.cmdSendandReceiveViaBulkTransfers.BackColor = System.Drawing.SystemColors.Control;
			this.cmdSendandReceiveViaBulkTransfers.Cursor = System.Windows.Forms.Cursors.Default;
			this.cmdSendandReceiveViaBulkTransfers.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cmdSendandReceiveViaBulkTransfers.ForeColor = System.Drawing.SystemColors.ControlText;
			this.cmdSendandReceiveViaBulkTransfers.Location = new System.Drawing.Point(256, 24);
			this.cmdSendandReceiveViaBulkTransfers.Name = "cmdSendandReceiveViaBulkTransfers";
			this.cmdSendandReceiveViaBulkTransfers.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.cmdSendandReceiveViaBulkTransfers.Size = new System.Drawing.Size(71, 36);
			this.cmdSendandReceiveViaBulkTransfers.TabIndex = 8;
			this.cmdSendandReceiveViaBulkTransfers.Text = "Send";
			this.cmdSendandReceiveViaBulkTransfers.UseVisualStyleBackColor = false;
			this.cmdSendandReceiveViaBulkTransfers.Click += new System.EventHandler(this.cmdSendAndReceiveViaBulkTransfers_Click);
			// 
			// fraBulkTransfer
			// 
			this.fraBulkTransfer.BackColor = System.Drawing.SystemColors.Control;
			this.fraBulkTransfer.Controls.Add(this.txtBulkDataToSend);
			this.fraBulkTransfer.Controls.Add(this.cmdSendandReceiveViaBulkTransfers);
			this.fraBulkTransfer.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.fraBulkTransfer.ForeColor = System.Drawing.SystemColors.ControlText;
			this.fraBulkTransfer.Location = new System.Drawing.Point(16, 139);
			this.fraBulkTransfer.Name = "fraBulkTransfer";
			this.fraBulkTransfer.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.fraBulkTransfer.Size = new System.Drawing.Size(341, 89);
			this.fraBulkTransfer.TabIndex = 4;
			this.fraBulkTransfer.TabStop = false;
			this.fraBulkTransfer.Text = "Bulk Transfers";
			// 
			// txtBulkDataToSend
			// 
			this.txtBulkDataToSend.AcceptsReturn = true;
			this.txtBulkDataToSend.BackColor = System.Drawing.SystemColors.Window;
			this.txtBulkDataToSend.Cursor = System.Windows.Forms.Cursors.IBeam;
			this.txtBulkDataToSend.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.txtBulkDataToSend.ForeColor = System.Drawing.SystemColors.WindowText;
			this.txtBulkDataToSend.Location = new System.Drawing.Point(16, 19);
			this.txtBulkDataToSend.MaxLength = 0;
			this.txtBulkDataToSend.Multiline = true;
			this.txtBulkDataToSend.Name = "txtBulkDataToSend";
			this.txtBulkDataToSend.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.txtBulkDataToSend.ScrollBars = System.Windows.Forms.ScrollBars.Both;
			this.txtBulkDataToSend.Size = new System.Drawing.Size(224, 59);
			this.txtBulkDataToSend.TabIndex = 5;
			this.txtBulkDataToSend.Text = "bulk transfer test data";
			// 
			// fraInterruptTransfer
			// 
			this.fraInterruptTransfer.BackColor = System.Drawing.SystemColors.Control;
			this.fraInterruptTransfer.Controls.Add(this.cmdSendAndReceiveViaInterruptTransfers);
			this.fraInterruptTransfer.Controls.Add(this.cboInterruptOutByte1);
			this.fraInterruptTransfer.Controls.Add(this.cboInterruptOutByte0);
			this.fraInterruptTransfer.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.fraInterruptTransfer.ForeColor = System.Drawing.SystemColors.ControlText;
			this.fraInterruptTransfer.Location = new System.Drawing.Point(379, 139);
			this.fraInterruptTransfer.Name = "fraInterruptTransfer";
			this.fraInterruptTransfer.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.fraInterruptTransfer.Size = new System.Drawing.Size(180, 89);
			this.fraInterruptTransfer.TabIndex = 1;
			this.fraInterruptTransfer.TabStop = false;
			this.fraInterruptTransfer.Text = "Interrupt Transfers";
			// 
			// cmdSendAndReceiveViaInterruptTransfers
			// 
			this.cmdSendAndReceiveViaInterruptTransfers.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cmdSendAndReceiveViaInterruptTransfers.Location = new System.Drawing.Point(98, 24);
			this.cmdSendAndReceiveViaInterruptTransfers.Name = "cmdSendAndReceiveViaInterruptTransfers";
			this.cmdSendAndReceiveViaInterruptTransfers.Size = new System.Drawing.Size(71, 36);
			this.cmdSendAndReceiveViaInterruptTransfers.TabIndex = 7;
			this.cmdSendAndReceiveViaInterruptTransfers.Text = "Send";
			this.cmdSendAndReceiveViaInterruptTransfers.UseVisualStyleBackColor = true;
			this.cmdSendAndReceiveViaInterruptTransfers.Click += new System.EventHandler(this.cmdSendAndReceiveViaInterruptTransfers_Click);
			// 
			// cboInterruptOutByte1
			// 
			this.cboInterruptOutByte1.BackColor = System.Drawing.SystemColors.Window;
			this.cboInterruptOutByte1.Cursor = System.Windows.Forms.Cursors.Default;
			this.cboInterruptOutByte1.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
			this.cboInterruptOutByte1.Font = new System.Drawing.Font("Arial", 8F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cboInterruptOutByte1.ForeColor = System.Drawing.SystemColors.WindowText;
			this.cboInterruptOutByte1.Location = new System.Drawing.Point(8, 56);
			this.cboInterruptOutByte1.Name = "cboInterruptOutByte1";
			this.cboInterruptOutByte1.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.cboInterruptOutByte1.Size = new System.Drawing.Size(74, 22);
			this.cboInterruptOutByte1.TabIndex = 3;
			// 
			// cboInterruptOutByte0
			// 
			this.cboInterruptOutByte0.BackColor = System.Drawing.SystemColors.Window;
			this.cboInterruptOutByte0.Cursor = System.Windows.Forms.Cursors.Default;
			this.cboInterruptOutByte0.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
			this.cboInterruptOutByte0.Font = new System.Drawing.Font("Arial", 8F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cboInterruptOutByte0.ForeColor = System.Drawing.SystemColors.WindowText;
			this.cboInterruptOutByte0.Location = new System.Drawing.Point(8, 24);
			this.cboInterruptOutByte0.Name = "cboInterruptOutByte0";
			this.cboInterruptOutByte0.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.cboInterruptOutByte0.Size = new System.Drawing.Size(72, 22);
			this.cboInterruptOutByte0.TabIndex = 2;
			// 
			// lstResults
			// 
			this.lstResults.BackColor = System.Drawing.SystemColors.Window;
			this.lstResults.Cursor = System.Windows.Forms.Cursors.Default;
			this.lstResults.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.lstResults.ForeColor = System.Drawing.SystemColors.WindowText;
			this.lstResults.HorizontalScrollbar = true;
			this.lstResults.ItemHeight = 16;
			this.lstResults.Location = new System.Drawing.Point(16, 248);
			this.lstResults.Name = "lstResults";
			this.lstResults.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.lstResults.Size = new System.Drawing.Size(543, 196);
			this.lstResults.TabIndex = 0;
			// 
			// cmdFindDevice
			// 
			this.cmdFindDevice.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cmdFindDevice.Location = new System.Drawing.Point(446, 44);
			this.cmdFindDevice.Name = "cmdFindDevice";
			this.cmdFindDevice.Size = new System.Drawing.Size(102, 59);
			this.cmdFindDevice.TabIndex = 11;
			this.cmdFindDevice.Text = "Find My Device";
			this.cmdFindDevice.Click += new System.EventHandler(this.cmdFindDevice_Click);
			// 
			// fraControWriteTransfer
			// 
			this.fraControWriteTransfer.BackColor = System.Drawing.SystemColors.Control;
			this.fraControWriteTransfer.Controls.Add(this.cboControlWriteByte1);
			this.fraControWriteTransfer.Controls.Add(this.cboControlWriteByte0);
			this.fraControWriteTransfer.Controls.Add(this.cmdControlWriteTransfer);
			this.fraControWriteTransfer.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.fraControWriteTransfer.ForeColor = System.Drawing.SystemColors.ControlText;
			this.fraControWriteTransfer.Location = new System.Drawing.Point(16, 25);
			this.fraControWriteTransfer.Name = "fraControWriteTransfer";
			this.fraControWriteTransfer.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.fraControWriteTransfer.Size = new System.Drawing.Size(193, 96);
			this.fraControWriteTransfer.TabIndex = 12;
			this.fraControWriteTransfer.TabStop = false;
			this.fraControWriteTransfer.Text = "Control Write Transfer";
			// 
			// cboControlWriteByte1
			// 
			this.cboControlWriteByte1.BackColor = System.Drawing.SystemColors.Window;
			this.cboControlWriteByte1.Cursor = System.Windows.Forms.Cursors.Default;
			this.cboControlWriteByte1.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
			this.cboControlWriteByte1.Font = new System.Drawing.Font("Arial", 8F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cboControlWriteByte1.ForeColor = System.Drawing.SystemColors.WindowText;
			this.cboControlWriteByte1.Location = new System.Drawing.Point(16, 58);
			this.cboControlWriteByte1.Name = "cboControlWriteByte1";
			this.cboControlWriteByte1.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.cboControlWriteByte1.Size = new System.Drawing.Size(67, 22);
			this.cboControlWriteByte1.TabIndex = 15;
			// 
			// cboControlWriteByte0
			// 
			this.cboControlWriteByte0.BackColor = System.Drawing.SystemColors.Window;
			this.cboControlWriteByte0.Cursor = System.Windows.Forms.Cursors.Default;
			this.cboControlWriteByte0.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
			this.cboControlWriteByte0.Font = new System.Drawing.Font("Arial", 8F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cboControlWriteByte0.ForeColor = System.Drawing.SystemColors.WindowText;
			this.cboControlWriteByte0.Location = new System.Drawing.Point(16, 30);
			this.cboControlWriteByte0.Name = "cboControlWriteByte0";
			this.cboControlWriteByte0.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.cboControlWriteByte0.Size = new System.Drawing.Size(67, 22);
			this.cboControlWriteByte0.TabIndex = 14;
			// 
			// cmdControlWriteTransfer
			// 
			this.cmdControlWriteTransfer.Font = new System.Drawing.Font("Arial", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.cmdControlWriteTransfer.Location = new System.Drawing.Point(102, 30);
			this.cmdControlWriteTransfer.Name = "cmdControlWriteTransfer";
			this.cmdControlWriteTransfer.Size = new System.Drawing.Size(71, 36);
			this.cmdControlWriteTransfer.TabIndex = 13;
			this.cmdControlWriteTransfer.Text = "Send";
			this.cmdControlWriteTransfer.UseVisualStyleBackColor = true;
			this.cmdControlWriteTransfer.Click += new System.EventHandler(this.cmdControlWriteTransfer_Click);
			// 
			// frmMain
			// 
			this.ClientSize = new System.Drawing.Size(579, 469);
			this.Controls.Add(this.fraControWriteTransfer);
			this.Controls.Add(this.cmdFindDevice);
			this.Controls.Add(this.fraControlReadTransfer);
			this.Controls.Add(this.fraBulkTransfer);
			this.Controls.Add(this.fraInterruptTransfer);
			this.Controls.Add(this.lstResults);
			this.Font = new System.Drawing.Font("Arial", 8F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
			this.Location = new System.Drawing.Point(21, 28);
			this.Name = "frmMain";
			this.StartPosition = System.Windows.Forms.FormStartPosition.Manual;
			this.Text = "WinUSB Demo";
			this.Load += new System.EventHandler(this.frmMain_Load);
			this.Closed += new System.EventHandler(this.frmMain_Closed);
			this.fraControlReadTransfer.ResumeLayout(false);
			this.fraBulkTransfer.ResumeLayout(false);
			this.fraBulkTransfer.PerformLayout();
			this.fraInterruptTransfer.ResumeLayout(false);
			this.fraControWriteTransfer.ResumeLayout(false);
			this.ResumeLayout(false);

		}

		#endregion

		//  This GUID must match the GUID in the device's INF file.
		//  To create a GUID in Visual Studio, click Tools > Create GUID.

		private const String WINUSB_DEMO_GUID_STRING = "{42CA71EC-CE1C-44c2-82DE-87D8D8FF6C1E}";

		private IntPtr deviceNotificationHandle;
		private Boolean myDeviceDetected = false;
		private DeviceManagement myDeviceManagement = new DeviceManagement();
		private String myDevicePathName;
		private WinUsbDevice myWinUsbDevice = new WinUsbDevice();

		internal frmMain frmMy;

		///  <summary>
		///  Define a class of delegates with the same parameters as 
		///  WinUsbDevice.ReadViaBulkTransfer and WinUsbDevice.ReadViaInterruptTransfer.
		///  Used for asynchronous reads from the device.
		///  </summary>

		private delegate void ReadFromDeviceDelegate
			(Byte pipeID,
			UInt32 bufferLength,
			ref Byte[] buffer,
			ref UInt32 lengthTransferred,
			ref Boolean success);

		///  <summary>
		///  Define a delegate with the same parameters as AccessForm.
		///  Used in accessing the application's form from a different thread.
		///  </summary>

		private delegate void MarshalToForm(String action, String textToAdd);

		///  <summary>
		///  Performs various application-specific functions that
		///  involve accessing the application's form.
		///  </summary>
		///  
		///  <param name="action"> A String that names the action to perform on the form. </param>
		///  <param name="formText"> Text to display on the form or an empty String. </param>
		///  
		/// <remarks>
		///  In asynchronous calls to WinUsb_ReadPipe, the callback function 
		///  uses this routine to access the application's form, which runs in 
		///  a different thread.
		/// </remarks>
		/// 
		private void AccessForm(String action, String formText)
		{
			try
			{
				//  Select an action to perform on the form:

				switch (action)
				{
					case "AddItemToListBox":

						lstResults.Items.Add(formText);

						break;
					case "AddItemToTextBox":

						txtBulkDataToSend.SelectedText = formText + "\r\n";

						break;
					case "EnableCmdSendandReceiveViaBulkTransfers":

						cmdSendandReceiveViaBulkTransfers.Enabled = true;
						cmdSendandReceiveViaBulkTransfers.Focus();

						break;
					case "EnableCmdSendandReceiveViaInterruptTransfers":

						cmdSendAndReceiveViaInterruptTransfers.Enabled = true;
						cmdSendAndReceiveViaInterruptTransfers.Focus();

						break;
					case "ScrollToBottomOfListBox":

						lstResults.SelectedIndex = lstResults.Items.Count - 1;

						break;

					default:
						break;
				}
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Display the device's speed in the list box.
		///  </summary>
		///  
		///  <remarks>
		///  Precondition: device speed was obtained by calling WinUsb_QueryDeviceInformation
		///  and stored in myDevInfo. 
		///  </remarks >

		private void DisplayDeviceSpeed()
		{
			String speed = "";

			myWinUsbDevice.QueryDeviceSpeed();

			try
			{
				switch (myWinUsbDevice.myDevInfo.devicespeed)
				{
					case 1:
						speed = "low";
						break;
					case 2:
						speed = "full";
						break;
					case 3:
						speed = "high";
						break;
				}

				lstResults.Items.Add("Device speed = " + speed);
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  If a device with the specified device interface GUID hasn't been previously detected,
		///  look for it. If found, open a handle to the device.
		///  </summary>
		///  
		///  <returns>
		///  True if the device is detected, False if not detected.
		///  </returns>

		private Boolean FindMyDevice()
		{
			Boolean deviceFound;
			String devicePathName = "";
			Boolean lastDevice;
			Boolean success;

			try
			{
				if (!(myDeviceDetected))
				{

					//  Convert the device interface GUID String to a GUID object: 

					System.Guid winUsbDemoGuid =
						new System.Guid(WINUSB_DEMO_GUID_STRING);

					// Fill an array with the device path names of all attached devices with matching GUIDs.

					deviceFound = myDeviceManagement.FindDeviceFromGuid
						(winUsbDemoGuid,
						ref devicePathName);

					if (deviceFound == true)
					{
						success = myWinUsbDevice.GetDeviceHandle(devicePathName);

						if (success)
						{
							lstResults.Items.Add("Device detected:");

							ScrollToBottomOfListBox();

							myDeviceDetected = true;

							// Save DevicePathName so OnDeviceChange() knows which name is my device.

							myDevicePathName = devicePathName;
						}
						else
						{
							// There was a problem in retrieving the information.

							myDeviceDetected = false;
							myWinUsbDevice.CloseDeviceHandle();
						}
					}

					if (myDeviceDetected)
					{

						// The device was detected.
						// Register to receive notifications if the device is removed or attached.

						success = myDeviceManagement.RegisterForDeviceNotifications
							(myDevicePathName,
							frmMy.Handle,
							winUsbDemoGuid,
							ref deviceNotificationHandle);

						if (success)
						{
							myWinUsbDevice.InitializeDevice();

							//Commented out due to unreliable response from WinUsb_QueryDeviceInformation.                            
							//DisplayDeviceSpeed(); 
						}
					}
					else
					{
						lstResults.Items.Add("Device not found.");
						cmdSendandReceiveViaBulkTransfers.Enabled = true;
						cmdSendAndReceiveViaInterruptTransfers.Enabled = true;
					}
				}
				else
				{
					lstResults.Items.Add("Device detected.");
				}

				ScrollToBottomOfListBox();

				return myDeviceDetected;

			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Retrieves received data from a bulk endpoint.
		///  This routine is called automatically when myWinUsbDevice.ReadViaBulkTransfer
		///  returns. The routine calls several marshaling routines to access the main form.       
		///  </summary>
		///  
		///  <param name="ar"> An object containing status information about the 
		///  asynchronous operation.</param>
		///  
		private void GetReceivedBulkData(IAsyncResult ar)
		{
			UInt32 bytesRead = 0;
			System.Text.ASCIIEncoding myEncoder = new System.Text.ASCIIEncoding();
			Byte[] receivedDataBuffer;
			String receivedtext = "";
			Boolean success = false;

			try
			{
				receivedDataBuffer = null;

				// Define a delegate using the IAsyncResult object.

				ReadFromDeviceDelegate deleg =
					((ReadFromDeviceDelegate)(ar.AsyncState));

				// Get the IAsyncResult object and the values of other paramaters that the
				// BeginInvoke method passed ByRef.

				deleg.EndInvoke
					(ref receivedDataBuffer,
					ref bytesRead,
					ref success, ar);

				// Display the received data in the form's list box.

				if ((ar.IsCompleted && success))
				{
					MyMarshalToForm("AddItemToListBox", "Data received via bulk transfer:");

					//  Convert the received bytes to a String for display.

					receivedtext = myEncoder.GetString(receivedDataBuffer);

					MyMarshalToForm("AddItemToListBox", receivedtext);
				}
				else
				{
					MyMarshalToForm("AddItemToListBox", "The attempt to read bulk data has failed.");
					myDeviceDetected = false;
				}

				MyMarshalToForm("ScrollToBottomOfListBox", "");

				// Enable requesting another transfer.

				MyMarshalToForm("EnableCmdSendandReceiveViaBulkTransfers", "");
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Retrieves received data from an interrupt endpoint.
		///  This routine is called automatically when myWinUsbDevice.ReadViaInterruptTransfer
		///  returns. The routine calls several marshaling routines to access the main form.
		///  Converts received bytes to hex strings for displaying.
		///  </summary>
		///  
		///  <param name="ar"> An object containing status information about the 
		///  asynchronous operation.</param>

		private void GetReceivedInterruptData(IAsyncResult ar)
		{
			String byteValue = "";
			UInt32 bytesRead = 0;
			Byte[] receivedDataBuffer = null;
			Boolean success = false;

			try
			{
				// Define a delegate using the IAsyncResult object.

				ReadFromDeviceDelegate deleg = ((ReadFromDeviceDelegate)(ar.AsyncState));

				// Get the IAsyncResult object and the values of other paramaters that the
				// BeginInvoke method passed ByRef.

				deleg.EndInvoke
					(ref receivedDataBuffer,
					ref bytesRead,
					ref success, ar);

				// Display the received data in the form's list box.

				if ((ar.IsCompleted && success))
				{
					MyMarshalToForm("AddItemToListBox", "Data received via interrupt transfer:");

					MyMarshalToForm("AddItemToListBox", " Received Data:");

					for (Int32 i = 0; i <= receivedDataBuffer.GetUpperBound(0); i++)
					{
						// Convert the byte value to a 2-character hex String.

						byteValue = String.Format("{0:X2} ", receivedDataBuffer[i]);

						MyMarshalToForm("AddItemToListBox", " " + byteValue);
					}
				}
				else
				{
					MyMarshalToForm("AddItemToListBox", "The attempt to read interrupt data has failed.");
					myDeviceDetected = false;
				}

				MyMarshalToForm("ScrollToBottomOfListBox", "");

				// Enable requesting another transfer.

				MyMarshalToForm("EnableCmdSendandReceiveViaInterruptTransfers", "");
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Initializes elements on the form.
		///  </summary>

		private void InitializeDisplay()
		{
			String byteValue;

			try
			{
				// Create dropdown list boxes.

				for (Int32 i = 0; i <= 255; i++)
				{
					byteValue = String.Format("{0:X2} ", i);

					frmMy.cboInterruptOutByte0.Items.Insert(i, byteValue);
					frmMy.cboInterruptOutByte1.Items.Insert(i, byteValue);
					frmMy.cboControlWriteByte0.Items.Insert(i, byteValue);
					frmMy.cboControlWriteByte1.Items.Insert(i, byteValue);
				}

				// Select a default value for each box

				frmMy.cboInterruptOutByte0.SelectedIndex = 0;
				frmMy.cboInterruptOutByte1.SelectedIndex = 128;
				frmMy.cboControlWriteByte0.SelectedIndex = 65;
				frmMy.cboControlWriteByte1.SelectedIndex = 53;

				if (!((myWinUsbDevice.IsWindowsXpOrLater())))
				{
					lstResults.Items.Add("The operating system is not Windows XP or later.");
					lstResults.Items.Add("The WinUsb driver requires Windows XP or later.");
				}
			}
			catch (Exception ex)
			{
				throw;
			}

		}

		///  <summary>
		///  Initiates a Control Read transfer. (Data stage is device to host.)
		///  </summary>
		///  
		private void InitiateControlInTransfer()
		{
			Boolean success;
			String formText = "";
			Byte[] dataStage = new Byte[2];

			try
			{
				// If the device hasn't been detected, was removed, or timed out on a previous attempt
				// to access it, look for the device.

				myDeviceDetected = FindMyDevice();

				if (myDeviceDetected)
				{
					success = myWinUsbDevice.Do_Control_Read_Transfer(ref dataStage);

					if (success)
					{
						formText = "Bytes received via Control Read transfer:";
						AccessForm("AddItemToListBox", formText);

						for (Int32 i = 0; i <= dataStage.Length - 1; i++)
						{
							formText = String.Format("{0:X2} ", dataStage[i]);
							AccessForm("AddItemToListBox", formText);
						}
					}
					else
					{
						formText = "Control Read transfer failed.";
						AccessForm("AddItemToListBox", formText);
						myDeviceDetected = false;
					}
				}
				ScrollToBottomOfListBox();
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Initiates a Control Write transfer. (Data stage is host to device.)
		///  Converts hex strings from combo boxes to byte values for sending.
		///  </summary>
		///  
		private void InitiateControlOutTransfer()
		{
			Boolean success;
			String comboBoxText = "";
			Byte[] dataStage = new Byte[2];
			String formText = "";

			try
			{
				// If the device hasn't been detected, was removed, or timed out on a previous attempt
				// to access it, look for the device.

				myDeviceDetected = FindMyDevice();

				if (myDeviceDetected)
				{
					// Get the value to send as a hex String.

					comboBoxText =
						System.Convert.ToString(cboControlWriteByte0.SelectedItem).TrimEnd(null);

					// Convert the String to a byte.

					dataStage[0] =
						Convert.ToByte(String.Format("{0:X2}", comboBoxText), 16);

					// Get the value to send as a hex String.

					comboBoxText =
						System.Convert.ToString(cboControlWriteByte1.SelectedItem).TrimEnd(null);

					// Convert the String to a byte.

					dataStage[1] =
						Convert.ToByte(String.Format("{0:X2}", comboBoxText), 16);

					success = myWinUsbDevice.Do_Control_Write_Transfer(dataStage);

					if (success)
					{
						formText = "Control Write transfer completed successfully.";
					}
					else
					{
						formText = "Control Write transfer failed.";
						myDeviceDetected = false;
					}

					AccessForm("AddItemToListBox", formText);
				}
				ScrollToBottomOfListBox();
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Enables accessing a form from another thread 
		///  </summary>
		///  
		///  <param name="action"> A String that names the action to perform on the form. </param>
		///  <param name="textToDisplay"> Text that the form displays or uses for 
		///  another purpose. Actions that don't use text ignore this parameter. </param>

		private void MyMarshalToForm(String action, String textToDisplay)
		{
			object[] args = { action, textToDisplay };
			MarshalToForm MarshalToFormDelegate = null;

			try
			{
				//  The AccessForm routine contains the code that accesses the form.

				MarshalToFormDelegate = new MarshalToForm(AccessForm);

				//  Execute AccessForm, passing the parameters in args.

				base.Invoke(MarshalToFormDelegate, args);
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Called when a WM_DEVICECHANGE message has arrived,
		///  indicating that a device has been attached or removed.
		///  </summary>
		///  
		///  <param name="m"> A message with information about the device. </param>

		internal void OnDeviceChange(Message m)
		{
			try
			{
				if ((m.WParam.ToInt32() == DeviceManagement.DBT_DEVICEARRIVAL))
				{

					//  If WParam contains DBT_DEVICEARRIVAL, a device has been attached.
					//  Find out if it's the device we're communicating with.

					if (myDeviceManagement.DeviceNameMatch(m, myDevicePathName))
					{
						lstResults.Items.Add("My device attached.");
					}

				}
				else if ((m.WParam.ToInt32() ==
					DeviceManagement.DBT_DEVICEREMOVECOMPLETE))
				{
					//  If WParam contains DBT_DEVICEREMOVAL, a device has been removed.
					//  Find out if it's the device we're communicating with.

					if (myDeviceManagement.DeviceNameMatch(m, myDevicePathName))
					{

						lstResults.Items.Add("My device removed.");

						//  Set MyDeviceDetected False so on the next data-transfer attempt,
						//  FindMyDevice() will be called to look for the device 
						//  and get a new handle.

						frmMy.myDeviceDetected = false;
					}
				}

				ScrollToBottomOfListBox();
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Initiates a read operation from a bulk IN endpoint.
		///  To enable reading without blocking the main thread, uses an asynchronous delegate.
		///  </summary>
		///  
		///  <remarks>
		///  To enable reading more than 64 bytes (with device firmware support), increase bytesToRead.
		///  </remarks> 

		private void ReadDataViaBulkTransfer()
		{

			IAsyncResult ar = null;
			Byte[] buffer = new Byte[64];
			UInt32 bytesRead = 0;
			UInt32 bytesToRead = System.Convert.ToUInt32(64);
			Boolean success = false;

			//  Define a delegate for the ReadViaBulkTransfer method of WinUsbDevice.

			ReadFromDeviceDelegate MyReadFromDeviceDelegate =
				new ReadFromDeviceDelegate(myWinUsbDevice.ReadViaBulkTransfer);

			try
			{
				//  The BeginInvoke method calls MyWinUsbDevice.ReadViaBulkTransfer to attempt 
				//  to read data. The method has the same parameters as ReadViaBulkTransfer,
				//  plus two additional parameters:
				//  GetReceivedBulkData is the callback routine that executes when 
				//  ReadViaBulkTransfer returns.
				//  MyReadFromDeviceDelegate is the asynchronous delegate object.

				ar = MyReadFromDeviceDelegate.BeginInvoke
					(System.Convert.ToByte(myWinUsbDevice.myDevInfo.bulkInPipe),
					bytesToRead,
					ref buffer,
					ref bytesRead,
					ref success,
					new AsyncCallback(GetReceivedBulkData),
					MyReadFromDeviceDelegate);
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Initiates a read operation from an interrupt IN endpoint.
		///  To enable reading without blocking the main thread, uses an asynchronous delegate.
		///  </summary>
		///  
		///  <remarks>
		///  To enable reading more than 2 bytes (with device firmware support), increase bytesToRead.
		///  </remarks>

		private void ReadDataViaInterruptTransfer()
		{
			IAsyncResult ar = null;
			Byte[] buffer = new Byte[2];
			UInt32 bytesRead = 0;
			UInt32 bytesToRead = System.Convert.ToUInt32(2);
			Boolean success = false;

			try
			{
				//  Define a delegate for the ReadViaInterruptTransfer method of WinUsbDevice.

				ReadFromDeviceDelegate MyReadFromDeviceDelegate = new ReadFromDeviceDelegate(myWinUsbDevice.ReadViaInterruptTransfer);

				//  The BeginInvoke method calls MyWinUsbDevice.ReadViaInterruptTransfer to attempt 
				//  to read data. The method has the same parameters as ReadViaInterruptTransfer,
				//  plus two additional parameters:
				//  GetReceivedInterruptData is the callback routine that executes when 
				//  ReadViaInterruptTransfer returns.
				//  MyReadFromDeviceDelegate is the asynchronous delegate object.

				ar = MyReadFromDeviceDelegate.BeginInvoke
					(System.Convert.ToByte(myWinUsbDevice.myDevInfo.interruptInPipe),
					bytesToRead,
					ref buffer,
					ref bytesRead,
					ref success,
					new AsyncCallback(GetReceivedInterruptData),
					MyReadFromDeviceDelegate);
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Scroll to the bottom of the list box and trim if needed.
		///  </summary>

		private void ScrollToBottomOfListBox()
		{
			try
			{
				lstResults.SelectedIndex = lstResults.Items.Count - 1;

				// If the list box is getting too large, trim its contents.

				if (lstResults.Items.Count > 1000)
				{
					for (Int32 i = 1; i <= 800; i++)
					{
						lstResults.Items.RemoveAt(i);
					}
				}
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Initiates sending data via a bulk transfer, then receiving data via a bulk transfer.
		///  </summary>

		private void SendAndReceiveViaBulkTransfers()
		{
			try
			{
				Boolean success;
				UInt32 bytesToSend;
				String dataToSend = "";
				Byte[] databuffer;
				String formText = "";
				System.Text.ASCIIEncoding myEncoder = new System.Text.ASCIIEncoding();

				//  Get data to send from the textbox.

				dataToSend = txtBulkDataToSend.Text;

				//  Convert the String to a byte array.

				databuffer = myEncoder.GetBytes(dataToSend);
				bytesToSend = Convert.ToUInt32(databuffer.Length);

				// If the device hasn't been detected, was removed, or timed out on a previous attempt
				// to access it, look for the device.

				myDeviceDetected = FindMyDevice();

				if (myDeviceDetected)
				{
					success = myWinUsbDevice.SendViaBulkTransfer
						(ref databuffer,
						bytesToSend);

					if (success)
					{
						formText = "Data sent via bulk transfer.";
					}
					else
					{
						formText = "Bulk OUT transfer failed.";
					}

					AccessForm("AddItemToListBox", formText);

					ReadDataViaBulkTransfer();
				}
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Initiates sending data via an interrupt transfer, then retrieving data 
		///  via an interrupt transfer.
		///  Converts hex strings from combo boxes to byte values for sending.
		///  </summary>

		private void SendAndReceiveViaInterruptTransfers()
		{
			Boolean success;
			UInt32 bytesToSend; //  winusb_readpipe requires this parameter to be a UINT32.
			String comboBoxText = "";
			Byte[] dataBuffer = new Byte[2];
			String formText = "";

			try
			{
				//  Get bytes to send from combo boxes.

				// Get the value to send as a hex String.

				comboBoxText =
					System.Convert.ToString(cboInterruptOutByte0.SelectedItem).TrimEnd(null);

				// Convert the String to a Byte.

				dataBuffer[0] =
					Convert.ToByte(String.Format("{0:X2}", comboBoxText), 16);

				// Get the value to send as a hex String.

				comboBoxText =
					System.Convert.ToString(cboInterruptOutByte1.SelectedItem).TrimEnd(null);

				// Convert the String to a Byte.

				dataBuffer[1] =
					Convert.ToByte(String.Format("{0:X2}", comboBoxText), 16);

				bytesToSend = Convert.ToUInt32(dataBuffer.Length);

				// If the device hasn't been detected, was removed, or timed out on a previous attempt
				// to access it, look for the device.

				myDeviceDetected = FindMyDevice();

				if (myDeviceDetected)
				{
					success = myWinUsbDevice.SendViaInterruptTransfer
						(ref dataBuffer,
						bytesToSend);

					if (success)
					{
						formText = "Data sent via interrupt transfer.";
					}
					else
					{
						formText = "Interrupt OUT transfer failed.";
					}

					AccessForm("AddItemToListBox", formText);
					ReadDataViaInterruptTransfer();
				}
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Perform actions that must execute when the program ends.
		///  </summary>

		private void Shutdown()
		{
			try
			{
				myWinUsbDevice.CloseDeviceHandle();

				myDeviceManagement.StopReceivingDeviceNotifications
					(deviceNotificationHandle);
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Perform actions that must execute when the program starts.
		///  </summary>

		private void Startup()
		{
			try
			{
				myWinUsbDevice = new WinUsbDevice();
				InitializeDisplay();
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Calls a routine to initiate a Control Read transfer. (Data stage is device to host.)
		///  </summary>

		private void cmdControlReadTransfer_Click(System.Object sender, System.EventArgs e)
		{
			InitiateControlInTransfer();
		}


		///  <summary>
		///  Calls a routine to initiate a Control Write transfer. (Data stage is host to device.)
		///  </summary>
		///   
		private void cmdControlWriteTransfer_Click(System.Object sender, System.EventArgs e)
		{
			InitiateControlOutTransfer();
		}

		///  <summary>
		///  Search for a specific device.
		///  </summary>

		private void cmdFindDevice_Click(System.Object sender, System.EventArgs e)
		{
			try
			{
				FindMyDevice();
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Calls a routine to exchange data via bulk transfers.
		///  </summary>

		private void cmdSendAndReceiveViaBulkTransfers_Click(System.Object eventSender, System.EventArgs eventArgs)
		{
			try
			{
				// Don't allow another transfer request until this one completes.

				cmdSendandReceiveViaBulkTransfers.Enabled = false;

				SendAndReceiveViaBulkTransfers();
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Calls a routine to exchange data via interrupt transfers.
		///  </summary>
		///  
		private void cmdSendAndReceiveViaInterruptTransfers_Click(System.Object sender, System.EventArgs e)
		{
			try
			{
				// Don't allow another transfer request until this one completes.

				cmdSendAndReceiveViaInterruptTransfers.Enabled = false;

				SendAndReceiveViaInterruptTransfers();
			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Perform shutdown operations.
		///  </summary>

		private void frmMain_Closed(System.Object eventSender, System.EventArgs eventArgs)
		{
			try
			{
				Shutdown();

			}
			catch (Exception ex)
			{
				throw;
			}
		}

		///  <summary>
		///  Perform startup operations.
		///  </summary>

		private void frmMain_Load(System.Object eventSender, System.EventArgs eventArgs)
		{
			try
			{
				frmMy = this;
				Startup();
			}
			catch (Exception ex)
			{
				throw;
			}

		}

		///  <summary>
		///  Overrides WndProc to enable checking for and handling
		///  WM_DEVICECHANGE messages.
		///  </summary>
		///  
		///  <param name="m"> A Windows message.
		///  </param> 
		///  
		protected override void WndProc(ref Message m)
		{
			try
			{
				// The OnDeviceChange routine processes WM_DEVICECHANGE messages.

				if (m.Msg == DeviceManagement.WM_DEVICECHANGE)
				{
					OnDeviceChange(m);
				}
				// Let the base form process the message.

				base.WndProc(ref m);
			}
			catch (Exception ex)
			{
				throw;
			}
		}
	}
}
